@@ -5,8 +5,17 @@ use strict;
 use warnings FATAL => 'all';
 use parent qw(Exporter);
 use Carp qw(croak);
+use Storable qw(freeze);
+use Algorithm::Diff qw(sdiff);
 
-our @EXPORT_OK = qw(diff dsplit dtraverse patch);
+$Storable::canonical = 1; # to have equal fingerprints for equal by data hashes
+
+our @EXPORT_OK = qw(
+    diff
+    dsplit
+    dtraverse
+    patch
+);
 
 sub _validate_meta($) {
     croak "Unsupported diff struct passed" if (ref $_[0] ne 'HASH');
@@ -14,7 +23,6 @@ sub _validate_meta($) {
         croak "Value for 'D' status must be hash or array"
             unless (ref $_[0]->{'D'} eq 'HASH' or ref $_[0]->{'D'} eq 'ARRAY');
     }
-    return 1;
 }
 
 =head1 NAME
@@ -23,11 +31,11 @@ Struct::Diff - Recursive diff tools for nested perl structures
 
 =head1 VERSION
 
-Version 0.70
+Version 0.81
 
 =cut
 
-our $VERSION = '0.70';
+our $VERSION = '0.81';
 
 =head1 SYNOPSIS
 
@@ -36,7 +44,7 @@ our $VERSION = '0.70';
     $a = {x => [7,{y => 4}]};
     $b = {x => [7,{y => 9}],z => 33};
 
-    $diff = diff($a, $b, noO => 1, noU => 1);       # omit unchanged and old values for changed items
+    $diff = diff($a, $b, noO => 1, noU => 1);       # omit unchanged items and old values for changed items
     # $diff == {D => {x => {D => [{I => 1,N => {y => 9}}]},z => {A => 33}}};
 
     $href = dsplit($diff);                          # divide diff
@@ -56,7 +64,7 @@ Nothing is exported by default.
 
 =head2 diff
 
-Returns HASH reference to recursive diff between two passed things. Beware when
+Returns hashref to recursive diff between two passed things. Beware when
 changing diff: some of it's substructures are links to original structures.
 
     $diff = diff($a, $b, %opts);
@@ -114,6 +122,11 @@ Drop removed item's data.
 
 =cut
 
+sub _freeze($) {
+    return $_[0] unless (ref $_[0]);
+    freeze($_[0]);
+}
+
 sub diff($$;@);
 sub diff($$;@) {
     my ($a, $b, %opts) = @_;
@@ -131,39 +144,34 @@ sub diff($$;@) {
         } else {
             $d->{'N'} = $b;
         }
-    } elsif ((ref $a eq 'ARRAY') and ($a ne $b)) {
-        my $s; # statuses collector
-        for (my $i = 0; $i < @{$a} and $i < @{$b}; $i++) {
-            my $tmp = diff($a->[$i], $b->[$i], %opts);
-            if (map { $s->{$_} = 1 } keys %{$tmp}) { # true if keys exists
-                $tmp->{'I'} = $i if ($hidden);
-                push @{$d->{'D'}}, $tmp;
-            } else {
-                $hidden = 1;
+    } elsif (ref $a eq 'ARRAY' and $a ne $b) {
+        my @sd = sdiff($a, $b, \&_freeze);
+        my $s; # status collector
+        for (my $i = 0; $i < @sd; $i++) {
+            my $item;
+            if ($sd[$i]->[0] eq 'u') {
+                $item->{'U'} = $sd[$i]->[1] unless ($opts{'noU'});
+            } elsif ($sd[$i]->[0] eq 'c') {
+                $item = diff($sd[$i]->[1], $sd[$i]->[2], %opts);
+            } elsif ($sd[$i]->[0] eq '+') {
+                $item->{'A'} = $sd[$i]->[2] unless ($opts{'noA'});
+            } else { # '-'
+                $item->{'R'} = $opts{'trimR'} ? undef : $sd[$i]->[1] unless ($opts{'noR'});
             }
-        }
-        if (@{$a} > @{$b}) {
-            if ($opts{'noR'}) {
+            unless ($item) {
                 $hidden = 1;
-            } else {
-                $s->{'R'} = 1;
-                map { push @{$d->{'D'}}, { 'R' => $opts{'trimR'} ? undef : $_ } } @{$a}[@{$b} .. $#{$a}];
-            }
-        }
-        if (@{$a} < @{$b}) {
-            if ($opts{'noA'}) {
-                $hidden = 1;
-            } else {
-                $s->{'A'} = 1;
-                map { push @{$d->{'D'}}, { 'A' => $_ } } @{$b}[@{$a} .. $#{$b}];
+                next;
             }
+            map { $s->{$_} = 1 } keys %{$item};
+            $item->{'I'} = $i if ($hidden);
+            push @{$d->{'D'}}, $item;
         }
 
         if ((my @k = keys %{$s}) == 1 and not ($hidden or exists $s->{'D'})) { # all have same status - return it
             map { $_ = $_->{$k[0]} } @{$d->{'D'}};
             $d->{$k[0]} = delete $d->{'D'};
         }
-    } elsif ((ref $a eq 'HASH') and ($a ne $b)) {
+    } elsif (ref $a eq 'HASH' and $a ne $b) {
         for my $key (keys %{{ %{$a}, %{$b} }}) { # go througth united uniq keys
             if (exists $a->{$key} and exists $b->{$key}) {
                 my $tmp = diff($a->{$key}, $b->{$key}, %opts);
@@ -263,18 +271,22 @@ Traverse through diff invoking callback function for subdiff statuses.
 
 =over 4
 
-=item callback
+=item depth E<lt>intE<gt>
+
+Don't dive deeper than defined number of levels
+
+=item callback E<lt>subE<gt>
 
 Mandatory option, must contain coderef to callback fuction. Four arguments will be passed to provided
 subroutine: value, path, status and ref to subdiff. Function must return some true value on success. Important:
 path (second argument) is actual for callback lifetime and will be immedeately changed afterwards.
 
-=item sortkeys
+=item sortkeys E<lt>subE<gt>
 
 Defines how will be traversed subdiffs for hashes. Keys will be picked randomely (depends on C<keys> behavior,
 default), sorted by provided subroutine (if value is a coderef) or lexically sorted if set to some other true value.
 
-=item statuses
+=item statuses E<lt>listE<gt>
 
 Exact list of statuses. Sequence defines invocation priority.
 
@@ -289,7 +301,7 @@ sub dtraverse($$;$) {
     croak "Statuses argument must be an arrayref" if ($o->{'statuses'} and ref $o->{'statuses'} ne 'ARRAY');
     _validate_meta($d);
 
-    if (exists $d->{'D'}) {
+    if (exists $d->{'D'} and (not exists $o->{'depth'} or $o->{'depth'} >= @{$p})) {
         if (ref $d->{'D'} eq 'ARRAY') {
             for (my $i = 0; $i < @{$d->{'D'}}; $i++) {
                 push @{$p}, [$i];
@@ -353,13 +365,11 @@ sub patch($$) {
     } elsif (exists $d->{'N'}) {
         ${$s} = $d->{'N'};
     }
-
-    return 1;
 }
 
 =head1 LIMITATIONS
 
-Struct::Diff fails on structures with loops in references. has_circular_ref from Data::Structure::Util can help
+Struct::Diff fails on structures with loops in references. has_circular_ref() from Data::Structure::Util can help
 to detect such structures.
 
 Only scalars, refs to scalars, ref to arrays and ref to hashes correctly traversed. All other data types compared
